% Lab 4
% Search in Prolog
% ===========================

% DATABASE
% ===========================
europe(lisbon, madrid, 503).
% europe(madrid, paris, 1053).
europe(dublin, london, 464).
europe(london, paris, 342).
europe(paris, brussels, 263).
europe(paris, luxembourg, 287).
europe(brussels, luxembourg, 187).
europe(brussels, amsterdam, 174).
europe(luxembourg, berlin, 600).
europe(amsterdam, berlin, 577).
europe(berlin, copenhagen, 355).
europe(berlin, warsaw, 516).
europe(berlin, prague, 280).
europe(berlin, vienna, 523).
europe(copenhagen, stockholm, 521).
europe(stockholm, helsinki, 396).
europe(helsinki, tallinn, 82).
europe(tallinn, riga, 280).
europe(riga, vilnius, 261).
europe(vilnius, warsaw, 393).
europe(prague, warsaw, 514).
europe(prague, bratislava, 291).
europe(prague, vienna, 250).
europe(bratislava, warsaw, 533).
europe(bratislava, budapest, 161).
europe(bratislava, vienna, 55).
europe(vienna, budapest, 214).
europe(vienna, ljubljana, 279).
europe(ljubljana, rome, 490).
europe(ljubljana, zagreb, 117).
europe(ljubljana, budapest, 380).
europe(budapest, zagreb, 299).
% europe(budapest, bucharest, 643).
europe(bucharest, sofia, 296).
europe(sofia, athens, 526).

way(X, Y, D) :- europe(X, Y, D).
way(X, Y, D) :- europe(Y, X, D).
% ===========================



% ===========================
% Task 2


%	dfs/4
% Search the map of Europe for a path from 'Start' to 'Goal' using depth-first search.
%
% Obtain a path along with its length.
%
% Use an auxiliary predicate dfs_aux/6 with the 
% 	5th argument being length accumulator; initialized to zero
%	6th argument being the closed list; initialized to contain the 'Start' city
dfs(Start, Goal, Path, Length) :- dfs_aux(Start, Goal, Path, Length, 0, [Start]).

dfs_aux(Goal, Goal, [Goal], LenAcc, LenAcc, _).
dfs_aux(Current, Goal, [Current|Path], Length, LenAcc, CL) :-
    way(Current, Child, Dist),
    \+ member(Child, CL),
    NewLenAcc is LenAcc + Dist,
    dfs_aux(Child, Goal, Path, Length, NewLenAcc, [Child|CL]).


%	dfs/5
% Search the map of Europe for a path from 'Start' to 'Goal' using depth-first search.
%
% Obtain a path (a journey) along with its reverse and its length.
%
% Use an auxiliary predicate dfs_aux/8 with the 
% 	6th argument being length accumulator; initialized to zero
%	7th argument being path reverse accumulator; initialized to to contain the 'Start' city
%	8th argument being the closed list; initialized to contain the 'Start' city
dfs(Start, Goal, Journey, Reverse, Length) :- 
    dfs_aux(Start, Goal, Journey, Reverse, Length, 0, [Start], [Start]).

dfs_aux(Goal, Goal, [Goal], RevAcc, LenAcc, LenAcc, RevAcc, _).
dfs_aux(Current, Goal, [Current|Path], Reverse, Length, LenAcc, RevAcc, CL) :-
    way(Current, Child, Dist),
    \+ member(Child, CL),
    NewLenAcc is LenAcc + Dist,
    dfs_aux(Child, Goal, Path, Reverse, Length, NewLenAcc, [Child|RevAcc], [Child|CL]).



% ===========================
% Task 4
compare_paths([L1|_], [L2|_]) :- L1 =< L2. 
% ?- findall([L, S, G], dfs(S, G, _, L), _Paths), max_member(compare_paths, Longest,_Paths).



% ===========================
% Task 5



%	bfs/4
% Search the map of Europe for a path from 'Start' to 'Goal' using breadth-first search.
%
% Obtain a path along with its length.
%
% Use an auxiliary predicate bfs_aux/4 with the 
% 	1st argument being the 'Goal' city
% 	2nd argument being the path
% 	3rd argument being the path's length
% 	4th argument being a queue such that
%		each element of the queue has the form 's(CurrentCity, LengthSoFar, PathSoFar)'
bfs(Start, Goal, Path, Length) :- bfs_aux(Goal, Path, Length, [s(Start, 0, [Start])]).

bfs_aux(Goal, Path, Length, [s(Goal, Length, Path)|_]).
bfs_aux(Goal, Path, Length, [s(Current, LenSoFar, PathSoFar)|Queue]) :-
    findall(s(Child, NewLen, [Child|PathSoFar]),
            (
            	way(Current, Child, Dist),
                \+ member(Child, PathSoFar),
                NewLen is LenSoFar + Dist
            ),
            Children
    ),
    append(Queue, Children, NewQueue),
    bfs_aux(Goal, Path, Length, NewQueue).



% ===========================
% Task 6


%	dfs_rb/4
% Search the map of Europe for a path from 'Start' to 'Goal' using depth-first search 
% and red-black tree to imlement the closed list.
%
% Obtain a path along with its length.
%
% Use an auxiliary predicate dfs_rb_aux/6 with the 
% 	5th argument being length accumulator; initialized to zero
%	6th argument being the closed list; initialized to a red black tree with one element
dfs_rb(Start, Goal, Path, Length) :- 
    list_to_rbtree([Start-0], RB),
    dfs_rb_aux(Start, Goal, Path, Length, 0, RB).

dfs_rb_aux(Goal, Goal, [Goal], LenAcc, LenAcc, _).
dfs_rb_aux(Current, Goal, [Current|Path], Length, LenAcc, RB) :-
    way(Current, Child, Dist),
    \+ rb_lookup(Child, 0, RB),
    NewLenAcc is LenAcc + Dist,
    rb_insert(RB, Child, 0, NewRB),
    dfs_rb_aux(Child, Goal, Path, Length, NewLenAcc, NewRB).



% ===========================
% Task 7


%	bfs_dl/4
% Search the map of Europe for a path from 'Start' to 'Goal' using breadth-first search,
% representing the queue using a difference list.
%
% Obtain a path along with its length.
%
% Use an auxiliary predicate bfs_dl_aux/4 with the 
% 	1st argument being the 'Goal' city
% 	2nd argument being the path
% 	3rd argument being the path's length
% 	4th argument being a queue such that
%		each element of the queue has the form 's(CurrentCity, LengthSoFar, PathSoFar)'
bfs_dl(Start, Goal, Path, Length) :- 
    Queue = [s(Start, 0, [Start])|X]-X,
    bfs_dl_aux(Goal, Path, Length, Queue).

bfs_dl_aux(Goal, Path, Length, Queue) :-
    \+ empty_dl(Queue),
    Queue = [s(Goal, Length, Path)|_]-_.
bfs_dl_aux(Goal, Path, Length, Queue) :-
    \+ empty_dl(Queue),
    Queue = [s(Current, LenSoFar, PathSoFar)|QueuePlus]-QueueMinus,
    findall(s(Child, NewLen, [Child|PathSoFar]),
            (
            	way(Current, Child, Dist),
                \+ member(Child, PathSoFar),
                NewLen is LenSoFar + Dist
            ),
            ChildrenPlus,
            ChildrenMinus
    ),
    append_dl(QueuePlus-QueueMinus, ChildrenPlus-ChildrenMinus, NewQueue),
    bfs_dl_aux(Goal, Path, Length, NewQueue).

append_dl(XPlus-YPlus, YPlus-YMinus, XPlus-YMinus).
empty_dl(X-Y) :- X == Y.

% Compare:
% ?- time((findall([L, S, G], bfs_dl(S, G, _, L), _Paths), length(_Paths, NumPaths), max_member(compare_paths, Longest,_Paths))).
% ?- time((findall([L, S, G], bfs(S, G, _, L), _Paths), length(_Paths, NumPaths), max_member(compare_paths, Longest,_Paths))).
